/*****************************
* ZeusCMD - OpenGL Base Code *
* By Grant James (ZEUS)      *
* http://www.zeuscmd.com     *
*****************************/
#include "glutil.h"

OpenGL *OpenGL::instance = NULL;

char *OpenGL::title = "";
int OpenGL::width = 640;
int OpenGL::height = 480;
int OpenGL::bits = 16;
bool OpenGL::fullscreen = false;

bool OpenGL::active = false;
bool OpenGL::finished = false;
bool OpenGL::keys[256];
int OpenGL::mouseX = 0;
int OpenGL::mouseY = 0;
bool OpenGL::mouseDownL = false;

HDC OpenGL::hdc = NULL;
HWND OpenGL::hwnd = NULL;
HGLRC OpenGL::hrc = NULL;

bool (*OpenGL::init)() = NULL;
void (*OpenGL::display)() = NULL;
void (*OpenGL::idle)() = NULL;
void (*OpenGL::resize)(int w, int h) = NULL;

bool OpenGL::initGL(char *t, int w, int h, bool full, int bpp)
{
	title = t;
	width = w;
	height = h;
	fullscreen = full;
	bits = bpp;

	HINSTANCE hInstance = GetModuleHandle(NULL);

	WNDCLASSEX wc;
	unsigned int pixelFormat;

	DWORD style;

	wc.cbSize = sizeof(wc);
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = (WNDPROC)WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(NULL, IDI_WINLOGO);
	wc.hIconSm = LoadIcon(NULL, IDI_WINLOGO);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = "ZeusCMD";

	if (!RegisterClassEx(&wc))
	{
		displayError("Registering Window Class");
		return false;
	}

	if (fullscreen)
	{
		DEVMODE displaySettings;
		memset(&displaySettings, 0, sizeof(displaySettings));
		displaySettings.dmSize = sizeof(displaySettings);
		displaySettings.dmPelsWidth = width;
		displaySettings.dmPelsHeight = height;
		displaySettings.dmBitsPerPel = bits;
		displaySettings.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;

		if (ChangeDisplaySettings(&displaySettings, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
		{
			displayError("Cannot change to requested display mode");
			fullscreen = false;
			killGL();
			return false;
		}
	}

	if (fullscreen)
	{
		style = WS_POPUP;
		ShowCursor(false);
	}
	else
	{
		style = WS_OVERLAPPEDWINDOW;
	}

	RECT rect;
	rect.left = 0; rect.top = 0;
	rect.right = width; rect.bottom = height;
	AdjustWindowRectEx(&rect, style, FALSE, 0);

	if (!(hwnd = CreateWindowEx(
		WS_EX_APPWINDOW,
		"ZeusCMD",
		title,
		WS_CLIPSIBLINGS | WS_CLIPCHILDREN | style,
		0, 0,
		rect.right - rect.left,
		rect.bottom - rect.top,
		NULL,
		NULL,
		hInstance,
		NULL)))
	{
		killGL();
		displayError("Creating Window");
		return false;
	}

	PIXELFORMATDESCRIPTOR pfd = 
	{
		sizeof(pfd),
			1,
			PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
			PFD_TYPE_RGBA,
			bits,
			0, 0, 0, 0, 0, 0,		// Color bits
			0, 0,					// Alpha bits
			0, 0, 0, 0, 0,			// Accumulation bits
			16,						// Depth buffer
			0,						// Stencil buffer
			0,						// Auxillary buffer
			PFD_MAIN_PLANE,
			0,						// reserved
			0, 0, 0					// Layer masks
	};

	if (!(hdc = GetDC(hwnd)))
	{
		killGL();
		displayError("Retrieving DC");
		return false;
	}

	if(!(pixelFormat = ChoosePixelFormat(hdc, &pfd)))
	{
		killGL();
		displayError("Choosing Pixel Format");
		return false;
	}

	if (!SetPixelFormat(hdc, pixelFormat, &pfd))
	{
		killGL();
		displayError("Setting Pixel Format");
		return false;
	}

	if (!(hrc = wglCreateContext(hdc)))
	{
		killGL();
		displayError("Creating Rendering Context");
		return false;
	}

	if (!wglMakeCurrent(hdc, hrc))
	{
		killGL();
		displayError("Activating Rendering Context");
		return false;
	}

	ShowWindow(hwnd, SW_SHOW);
	SetForegroundWindow(hwnd);
	SetFocus(hwnd);

	if (resize)
		resize(width, height);

	if (init)
		return init();

	return true;
}

bool OpenGL::changeDisplaySettings(int w, int h, bool full, int bpp)
{
	killGL();

	int oldWidth = width;
	int oldHeight = height;
	bool oldFull = fullscreen;
	int oldBpp = bits;

	if (!initGL(title, w, h, full, bpp))
	{
		initGL(title, oldWidth, oldHeight, oldFull, oldBpp);
		return false;
	}

	if (init)
		return init();

	return true;
}

void OpenGL::killGL()
{
	if (fullscreen)
	{
		ChangeDisplaySettings(NULL, 0);
		ShowCursor(true);
	}

	if (hrc)
	{
		if (!wglMakeCurrent(NULL, NULL))
			displayError("Deactivating Rendering Context");

		if (!wglDeleteContext(hrc))
			displayError("Deleting Rendering Context");

		hrc = NULL;
	}

	if (hdc && !ReleaseDC(hwnd, hdc))
		displayError("Releasing Device Context");

	hdc = NULL;

	if (hwnd && !DestroyWindow(hwnd))
		displayError("Destroying Window");

	hwnd = NULL;

	if (!UnregisterClass("ZeusCMD", GetModuleHandle(NULL)))
		displayError("Unregistering Class");
}

LRESULT OpenGL::WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{

	case WM_ACTIVATE :
		{
			active = HIWORD(wParam) ? false : true;

			return 0;
		}

	case WM_CLOSE :
		{
			PostQuitMessage(0);
			return 0;
		}

	case WM_SIZE :
		{
			width = LOWORD(lParam);
			height = HIWORD(lParam);

			if (resize)
				resize(LOWORD(lParam), HIWORD(lParam));

			return 0;
		}

	case WM_KEYDOWN :
		{
			keys[wParam] = true;
			return 0;
		}

	case WM_KEYUP :
		{
			keys[wParam] = false;
			return 0;
		}

	case WM_MOUSEMOVE :
		{
			mouseX = LOWORD(lParam);
			mouseY = height - (HIWORD(lParam) + 1);

			return 0;
		}

	case WM_LBUTTONDOWN : mouseDownL = true; return 0;
	case WM_LBUTTONUP : mouseDownL = false; return 0;

	}

	return DefWindowProc(hwnd, msg, wParam, lParam);
}

void OpenGL::displayError(char *title)
{
	LPVOID lpMsgBuf;

	FormatMessage( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM | 
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		GetLastError(),
		0,
		(LPTSTR) &lpMsgBuf,
		0,
		NULL 
		);

	// Display the string.
	MessageBox( NULL, (LPCTSTR)lpMsgBuf, title, MB_OK | MB_ICONERROR);

	// Free the buffer.
	LocalFree( lpMsgBuf );
}

void OpenGL::runMessageLoop()
{
	MSG msg;

	while(!finished)
	{
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if (msg.message == WM_QUIT)
				break;

			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		if (active)
		{
			if (idle)
			{
				idle();

				if (keys[VK_F1])
				{
					keys[VK_F1] = false;

					fullscreen = !fullscreen;

					if (fullscreen)
						fullscreen = this->changeDisplaySettings(1024, 768, true, 32);
					else
						this->changeDisplaySettings(500, 500, false, 32);
				}
				else if (keys[VK_ESCAPE])
					this->endMessageLoop();
			}

			display();
			SwapBuffers(hdc);
		}
	}

	killGL();
}
